/*
 * Copyright (c) 2018  Fabien Feschet <fabien.feschet@dg-medical.fr>
 *
 * All rights reserved.
 *
 * Licence   : Simplified BSD License see LICENSE file.
 *
 */

/**
 * @file hosvd2.cpp
 * @brief HOSVD denoising functions
 *
 * @authors Fabien Feschet <fabien.feschet@dg-medical.fr>
 *
 **/

#include <iostream>
#include <algorithm>
#include <cmath>

#include "hosvd2.h"

#ifdef _OPENMP
#include <omp.h>
#endif

/**
 * @brief [Parallel] HOSVD denoising with image subdivision
 *
 * @param sigma:                noise level
 * @param img_clean:            clean image for Oracle denoising
 * @param img_noisy:            noisy image (input of the function)
 * @param img_denoised:         denoised image (output of the function)
 * @param img_denoised_wiener:  denoised image using Wiener step (output of the function)
 * @param img_denoised_oracle:  denoised image using Oracle denoising (output of the function)
 * @param width, height, chnls: input image characteristics
 * @param color_space:          color space for computation
 * @param out:                  settings to enable Wiener and Oracle denoising
 *                              default is to only performed HOSVD hard thresholding
 * @param settings:             parameters settings for experimental purposes
 *
 * @return EXIT_FAILURE in case of error, EXIT_SUCCESS otherwise
 **/
int hosvd(
    const float sigma
,   std::vector<float> &img_clean
,   std::vector<float> &img_noisy
,   std::vector<float> &img_denoised
,   std::vector<float> &img_denoised_wiener
,   std::vector<float> &img_denoised_oracle
,   const unsigned width
,   const unsigned height
,   const unsigned chnls
,   const unsigned color_space
,   Output_params &out
,   Options_params &settings
        ) {
    //! Parameters initialization
    unsigned kHard = settings.kHard;
    unsigned pHard = settings.pHard;
    unsigned P = 2*settings.nHard + 1;
    unsigned Q = P; // This code is for rectangular search window so we force squared one
    unsigned K = settings.K;
    unsigned nbThreads = settings.nbThreads;
    std::vector<float> img_null;

    //! Transformation from RGB to color space
    if (color_space_transform(img_noisy, color_space, width, height, chnls, true)
            != EXIT_SUCCESS)
        return EXIT_FAILURE;

    //! 1st Pass denoising
    std::cout << "HOSVD 1st pass start:" << std::flush;
    hosvd_sub(sigma,img_noisy,img_denoised,false,img_null,false,img_null,
              width,height,chnls,color_space,P,Q,K,pHard,kHard,kHard,settings.thresholdHD,settings.thresholdBM,nbThreads);
    std::cout << " done" << std::endl;

    //! 2nd pass - Wiener if enabled
    if (out.hasOutputDenoised2) {
            std::cout << "HOSVD 2nd pass start: Wiener filtering..." << std::flush;
            // Denoised with Wiener filtering enabled
            hosvd_sub(sigma,img_noisy,img_denoised_wiener,true,img_denoised,false,img_null,
                      width,height,chnls,color_space,P,Q,K,pHard,kHard,kHard,settings.thresholdHD,settings.thresholdBM,nbThreads);
            std::cout << " done" << std::endl;
        }
    //! Oracle Denoising if enabled
    if (out.hasOutputOracle) {
            // Denoised with Oracle filtering enabled
            std::cout << "HOSVD Oracle start:" << std::flush;
            hosvd_sub(sigma,img_noisy,img_denoised_oracle,false,img_null,true,img_clean,
                      width,height,chnls,color_space,P,Q,K,pHard,kHard,kHard,settings.thresholdHD,settings.thresholdBM,nbThreads);
            std::cout << " done" << std::endl;
        }

    //! Inverse color space transform to RGB for any computed output images
    if (out.hasOutputOracle) {
            if (color_space_transform(img_denoised_oracle, color_space, width, height, chnls, false)
                    != EXIT_SUCCESS) return EXIT_FAILURE;
        }
    if (out.hasOutputDenoised2) {
            if (color_space_transform(img_denoised_wiener, color_space, width, height, chnls, false)
                    != EXIT_SUCCESS) return EXIT_FAILURE;
        }
    if (color_space_transform(img_denoised, color_space, width, height, chnls, false)
            != EXIT_SUCCESS) return EXIT_FAILURE;
    if (color_space_transform(img_noisy, color_space, width, height, chnls, false)
            != EXIT_SUCCESS) return EXIT_FAILURE;
    //! End
    return EXIT_SUCCESS;
}

/**
 * @brief Parallel HOSVD denoising with image subdivision
 * @brief All input images are supposed to be symmetrized before calling
 *
 * @param sigma:                noise level
 * @param img_noisy:            noisy image (input of the function)
 * @param img_denoised:         denoised image (output of the function)
 * @param doWiener, img_wiener  Wiener filtering using a previously denoised image (input)
 * @param doOracle, img_oracle  Oracle filtering using a clean image (input)
 * @param width, height, chnls: input image characteristics
 * @param color_space:          color space for the computation
 * @param P, Q:                 total size of the neighborhood for searching similar patches
 * @param K:                    total number of max similar patches allowed
 * @param pHW:                  acceleration factor
 * @param p,q:                  Size of patches is pxq
 * @param thresholdHD:          multiplier for Hard Threshold (shoudl be 1.0 for no effect)
 * @param thresholdBM:          multiplier for Block Matching (shoudl be 1.0 for no effect)
 * @param nb_threads:           max number of used threads
 *
 * @return none because img_denoised is modified
 **/
void hosvd_sub(const float sigma, const std::vector<float> &img_noisy, std::vector<float> &img_denoised, 
               const bool doWiener, const std::vector<float> &img_wiener,
               const bool doOracle, const std::vector<float> &img_oracle,
               const unsigned width, const unsigned height, const unsigned chnls, const unsigned color_space,
               const unsigned P, const unsigned Q,
               const unsigned K,
               const unsigned pHW,
               const unsigned p, const unsigned q,
               const float thresholdHD,
               const float thresholdBM,
               unsigned nb_threads)
{
    //! Estimation of sigma on each channel this depends on the color_space
    std::vector<float> sigma_table(chnls);
    if (estimate_sigma(sigma, sigma_table, chnls, color_space) != EXIT_SUCCESS) {
        std::cout << "Unable to fix noise in the required color space, force using "
                  << sigma << " in each channel." << std::endl;
        for (auto it = sigma_table.begin(); it != sigma_table.end(); ++it)
            *it = sigma;
    } 
    //! Check allocation memory and fill with 0.0f
    if (img_denoised.size() != img_noisy.size())
        img_denoised.resize(img_noisy.size());
    std::fill(img_denoised.begin(), img_denoised.end(), 0.0f);
    //! For aggregation part
    std::vector<float> denominator(img_denoised.size(), 0.0f);
    // Symmetrization of input image(s) for symmetric block matching or usage
    std::vector<float> img_noisy_sym;
    symmetrize(img_noisy,img_noisy_sym,width,height,chnls,P,Q);
    std::vector<float> img_wiener_sym;
    if (doWiener)
        symmetrize(img_wiener,img_wiener_sym,width,height,chnls,P,Q);
    std::vector<float> img_oracle_sym;
    if (doOracle)
        symmetrize(img_oracle,img_oracle_sym,width,height,chnls,P,Q);
    //! Precompute Bloc-Matching
    std::vector<std::vector<std::pair<float, unsigned int>>> patch_table;
    precompute_BM(patch_table // output table for each patch
                  ,(doWiener ? img_wiener_sym : img_noisy_sym), width, height, chnls // input symetrized image
                  ,p, q // Size pxp for patches
                  ,K // Max number of other elements in output list for each input patch, the input patch is in the list
                  ,P, Q // Size PxQ for the proximity of patches
                  ,thresholdBM // Threshold value to keep or not a similar patch
                  ,sigma_table
                  ,nb_threads);
    //! Prepare parallel computation
    Tiling_2d decomposition;
    int status = EXIT_FAILURE;
    status = load_balancing_2d(nb_threads, width, height, decomposition);
    if (status == EXIT_FAILURE)
        return;
    //! Create nb_threads storages for output subimages data
    std::vector< std::vector< float > > img_denoised_sub(nb_threads);
    std::vector< std::vector< float > > denominator_sub(nb_threads);
    for (auto s = 0u ; s < nb_threads ; ++s) {
        unsigned xmin, xmax, ymin, ymax;
        std::tie(xmin, xmax, ymin, ymax) = decomposition[s];
        img_denoised_sub[s].resize((xmax-xmin+2*P)*(ymax-ymin+2*Q)*chnls);
        denominator_sub[s].resize(img_denoised_sub[s].size());
    }
#ifdef _OPENMP
    omp_set_dynamic(0); omp_set_num_threads(nb_threads);
#endif
#pragma omp parallel shared(img_noisy_sym, img_wiener_sym, img_oracle_sym, patch_table, decomposition, img_denoised_sub, denominator_sub)
{
#pragma omp for schedule(dynamic) nowait
    for (auto s = 0u; s < nb_threads; ++s)
    {
        // Initialization
        const float weightCore = 1.0f;
        double lambdaHard3D = 0.0; //! dynamic Threshold for Hard Thresholding in HOSVD
        const auto row= width+2*P; const auto col= height+2*Q; // Correspond to noisy symmetric image
        const auto dec= row*col;                         // idem
        // Get region of support
        unsigned xmin, xmax, ymin, ymax;
        std::tie(xmin, xmax, ymin, ymax) = decomposition[s];
        const auto w = xmax-xmin+2*P;
        const auto h = ymax-ymin+2*Q;
        const auto dec_out = w*h;
        // Prepare loop
        std::vector<float>::const_iterator it_in_ref = img_noisy_sym.cbegin();
        if (doWiener)
            it_in_ref = img_wiener_sym.cbegin();
        if (doOracle)
            it_in_ref = img_oracle_sym.cbegin();
        auto it_in = it_in_ref;
        auto it_out = img_denoised_sub[s].begin();
        auto it_out_ref = it_out;
        // Sequential loop over the support
        for (auto patch_y = ymin; patch_y < ymax; patch_y+=pHW)
        {
            for (auto patch_x = xmin; patch_x < xmax; patch_x+=pHW)
            {
                //! Number of similar patches
                const unsigned k_r = patch_x+width*patch_y;
                const unsigned nSx_r = patch_table[k_r].size();
                //! Build of the nD group: 3D for chnls == 1 and 4D for chnls==3
                Eigen::MatrixXd unfold1(p, q * nSx_r * chnls);
                Eigen::MatrixXd unfold1_noisy(p, q * nSx_r * chnls);
                Eigen::MatrixXd unfold2(q, p * nSx_r * chnls);
                Eigen::MatrixXd unfold3(chnls, p * q * nSx_r); // Useless if chnls==1
                Eigen::MatrixXd unfold4(nSx_r, p * q * chnls);
                for (auto t = 0u; t < nSx_r; ++t)
                {
                    auto ind = patch_table[k_r][t].second;
                    for (auto z = 0u; z < chnls; ++z)
                    {
                        it_in = it_in_ref + ind;
                        for (auto y = 0u; y < q; ++y)
                        {
                            for (auto x = 0u; x < p; ++x)
                            {
                                float val = *it_in;
                                unfold1(x, y + z * q + t * q * chnls) = val;
                                unfold2(y, x + z * p + t * p * chnls) = val;
                                unfold3(z, x + y * p + t * p * q) = val;
                                unfold4(t, x + y * p + z * p * q) = val;
                                if (doWiener || doOracle)
                                    unfold1_noisy(x, y + z * q + t * q * chnls) = img_noisy_sym[it_in - it_in_ref];
                                ++it_in;
                            }
                            it_in += row - p;
                        }
                        ind += dec;
                    }
                }
                //! 3 or 4 SVD
                Eigen::JacobiSVD<Eigen::MatrixXd> svd1(unfold1,Eigen::ComputeFullU);
                Eigen::JacobiSVD<Eigen::MatrixXd> svd2(unfold2,Eigen::ComputeFullU);
                Eigen::JacobiSVD<Eigen::MatrixXd> svd3(unfold3,Eigen::ComputeFullU);
                Eigen::JacobiSVD<Eigen::MatrixXd> svd4(unfold4,Eigen::ComputeFullU);
                //! X1: unfolding 1 of the core of the HOSVD
                //! X1 = U1^T unfold1 ( U4 x U2)
                //! x is the Kronecker product ( voir CICHOCKI arxiv 1403.4462 )
                Eigen::MatrixXd U42;
                if (chnls == 1)
                    U42 = Eigen::kroneckerProduct(svd4.matrixU(), svd2.matrixU());
                if (chnls == 3)
                    U42 = Eigen::kroneckerProduct(svd4.matrixU(), Eigen::kroneckerProduct(svd3.matrixU(),svd2.matrixU()));
                Eigen::MatrixXd core, coreA;
                if (doOracle)
                    core = svd1.matrixU().transpose() * unfold1_noisy * U42;
                else
                    core = svd1.matrixU().transpose() * unfold1 * U42;
                if (doWiener) {
                    //! Perform Wiener construction: Core = CoreA.*(core.^2./(core.^2 + sigma^2))
                    coreA = svd1.matrixU().transpose() * unfold1_noisy * U42;
                    double st2; double val;
                    for (auto z = 0u; z < chnls; ++z) {
                        st2 = sigma_table[z]*sigma_table[z];
                        for (auto t = 0u; t < nSx_r; ++t) {
                            for (auto y = 0u; y < q ; ++y)
                                for (auto x = 0u; x < p ; ++x) {                                    
                                    val = core(x,y + z * q + t * q * chnls);
                                    val = val * val;
                                    val = val / (val + st2);
                                    val = coreA(x,y + z * q + t * q * chnls) * val;
                                    core(x,y + z * q + t * q * chnls) = val;
                                }
                        }
                    }
                }  else {
                    //! Thresholding value
                    lambdaHard3D = sqrt(2*log(p*q*nSx_r*chnls))*thresholdHD;
                    //! Threshold the whole core
                    double val, thresh;
                    for (auto z = 0u; z < chnls; ++z) {
                        thresh = lambdaHard3D * sigma_table[z];
                        for (auto t = 0u; t < nSx_r; ++t) {
                            for (auto y = 0u; y < q ; ++y)
                                for (auto x = 0u; x < p ; ++x) {
                                    val = core(x,y + z * q + t * q * chnls);
                                    // hard-thresholding
                                    if (static_cast<float>(fabs(val)) <= thresh)
                                        core(x,y + z * q + t * q * chnls) = 0.0f;
                                }
                        }
                    }
                }
                //! Reconstruction of the 1 unfolding of the tensor
                //! A1 = U1 Thres(X1) (U4^T x U2^T)
                Eigen::MatrixXd fromCore = svd1.matrixU() * core * U42.transpose();
                //! Averaging done in the output sub image
                for (auto t = 0u; t < nSx_r; ++t)
                {
                    // First point of the patch in noisy extended image
                    auto ind = patch_table[k_r][t].second;
                    // Absolute coordinates in image input which has been symmetrized -> ind_x, ind_y
                    unsigned ind_x = (ind % row); unsigned ind_y = (ind / row);
                    // Coordinates in the local sub image
                    ind_x -= xmin;  ind_y -= ymin;
                    auto ref = ind_x + w * ind_y;
                    unsigned delta;
                    for (auto z = 0u; z < chnls; ++z)
                    {
                        it_out = it_out_ref + ref;
                        delta = ref;
                        for (auto y = 0u; y < q; ++y)
                        {
                            for (auto x = 0u; x < p; ++x)
                            {
                                (*it_out) += fromCore(x, y + z * q + t * q * chnls) * weightCore;
                                denominator_sub[s][delta] += weightCore;
                                ++it_out;++delta;
                            }
                            it_out += w - p; delta += w - p;
                        }
                        // Next
                        ref += dec_out;
                    }
                }
            }
        }
        // Compute the output sub images
        for (auto k = 0u; k < img_denoised_sub[s].size(); ++k)
        {
            if (denominator_sub[s][k] == 0.0f)
                img_denoised_sub[s][k] = 0.0f;
            else
            {
                img_denoised_sub[s][k] /= denominator_sub[s][k];
                denominator_sub[s][k] = 1.0f;
            }
        }
    }
}
    //! Final reconstruction to avoid division by 0.0f on the outer boundary
    auto dec_out = width*height;
    for (auto s = 0u ; s < nb_threads ; ++s)
    {
        // Get region of support
        unsigned xmin, xmax, ymin, ymax;
        std::tie(xmin, xmax, ymin, ymax) = decomposition[s];
        auto w = xmax-xmin+2*P;
        auto h = ymax-ymin+2*Q;
        auto dec_in = w*h;
        // Intersection between region with boundaries and output support
        auto Xmin = (xmin < P ? 0 : xmin-P); auto Ymin = (ymin < Q ? 0 : ymin-Q);
        auto Xmax = (xmax+P>=width ? width : xmax+P); auto Ymax = (ymax+Q>=height ? height : ymax+Q);
        auto W = Xmax-Xmin;
        auto delta_x = w-W; // To complete one line of input
        // Local coordinates in input
        auto local_x = P+Xmin-xmin; auto local_y = Q+Ymin-ymin;
        // Output iterators
        auto it_denoised = img_denoised.begin()+Xmin+Ymin*width;
        auto it_out = it_denoised;
        auto ref_out = it_out - it_denoised;
        auto it_denominator = denominator.begin() + Xmin+Ymin*width;
        auto it_out_denom = it_denominator;
        // Input iterators
        auto it_sub = img_denoised_sub[s].begin()+local_x+local_y*w;
        auto it_in  = it_sub;
        auto ref_in = it_in - it_sub;
        auto it_sub_denominator = denominator_sub[s].begin()+local_x+local_y*w;
        auto it_in_denom = it_sub_denominator;
        // Save value and denominator
        for (auto z = 0u ; z < chnls ; ++z)
        {
            it_out = it_denoised + ref_out; it_out_denom = it_denominator + ref_out;
            it_in = it_sub + ref_in; it_in_denom = it_sub_denominator + ref_in;
            for (auto y = Ymin ; y < Ymax ; ++y)
            {
                for (auto x = Xmin ; x < Xmax ; ++x)
                {
                    *it_out += *it_in; ++it_out; ++it_in;
                    *it_out_denom += *it_in_denom; ++it_out_denom; ++it_in_denom;
                }
                it_out += width-W; it_out_denom += width-W;
                it_in += delta_x; it_in_denom += delta_x;
            }
            ref_in  += dec_in;
            ref_out += dec_out;
        }
    }
    // Last mean
    auto it_im = img_denoised.begin();
    auto it_denom = denominator.cbegin();
    auto it_im_noisy = img_noisy.cbegin();
    while (it_im != img_denoised.end())
    {
        if (*it_denom == 0.0f) *it_im = *it_im_noisy;
        else *it_im /= (*it_denom);
        // Rounding
        *it_im = static_cast<float>(floor((*it_im) + 0.5f));
        ++it_im; ++ it_denom; ++it_im_noisy;
    }
}

/** ---------------------------------- **/
/** - Preprocessing / Postprocessing - **/
/** ---------------------------------- **/

/**
 * @brief Bounded set management insert
 *
 * @param set:   set to manage
 * @param K:     bounded size of the set
 * @param value: value to insert
 *
 * @return none because set is modified
 **/
void insert_boundedSet(const std::pair<float, unsigned int> &value,
                       std::vector<std::pair<float, unsigned int>> &set,
                       const unsigned int K)
{
    // Add value to the set whose size is bounded
    // three cases: less than K-1   values => simply insert
    //              exactly   K-1   values => insert then make_heap to have the max at top
    //              exactly   K     values => just pop_head and push_head
    auto size = set.size();
    if (size < K)
    {
        set.push_back(value);
    }
    if (size == K-1)
    {
        // make a heap for future insert
        std::make_heap(set.begin(), set.end(), greaters_similarity<float, unsigned int>());
    }
    if (size == K)
    { // Special insertion to keep K+1 patches
        // old max value: we now that the vector is indeed a heap
        auto max_value = set.front().first;
        // if new patch is closer than old max patch then add new patch to the heap
        if (max_value > value.first)
        {
            // switch position of max
            std::pop_heap(set.begin(), set.end());
            // supress max distance patch
            set.pop_back();
            // insert current distance
            set.push_back(value);
            std::push_heap(set.begin(), set.end(), greaters_similarity<float, unsigned int>());
        }
    }
}

/**
 * @brief Inter-patches similarity: L2 distance sum over all channels
 *
 * @param P1, P2: const_iterator in the image
 * @param width, height, chnls: size of img and number of channels
 * @param p,q: size of patches
 * @param P,Q: boundary of symetrization in the image
 *
 * @return float similarity
 **/
float compute_dist(std::vector<float>::const_iterator P1,
                   std::vector<float>::const_iterator P2,
                   const unsigned int p, const unsigned int q,
                   const unsigned width, const unsigned height, const unsigned chnls,
                   const unsigned P, const unsigned Q)
{
    // This function is for kernel L2: sum over channel
    // Hypothesis: P1 and P2 are set on the first channel !
    auto chnls_next = (width + 2 * P) * (height + 2 * Q);
    unsigned int displQ = width + 2 * P - p;
    auto P1_chnls = P1;
    auto P2_chnls = P2;
    float similarity = 0.0f;
    for (auto c = 0u; c < chnls ; ++c)
    {
        P1 = P1_chnls;
        P2 = P2_chnls;
        for (auto y = 0u; y < q; ++y)
        {
            for (auto x = 0u; x < p; ++x)
            {
                similarity += ((*P1) - (*P2)) * ((*P1) - (*P2));
                ++P1;
                ++P2;
            }
            P1 += displQ;
            P2 += displQ;
        }
        P1_chnls += chnls_next;
        P2_chnls += chnls_next;
    }
    return similarity;
}

/**
 * @brief Precompute Bloc Matching (distance inter-patches) by summing over all channels
 *
 * @param patch_table: for each patch in the image, will contain all coordinates of its similar patches
 * @param img: image on which the distance is computed
 * @param width, height, chnls: size of img and number of channels
 * @param p,q: size of patch
 * @param K: maximum similar patches wanted
 * @param P,Q: size of the boundary of img
 * @param thresholdBM: initial threshold used to determinate similarity between patches
 *                     will be multiplied by the norm of sigma
 * @param sigma_table: noise in each channel
 * @param nb_threads : max number of used threads
 *
 * @return none since patch_table is modified.
 **/
void precompute_BM(std::vector<std::vector<std::pair<float, unsigned int>>> &patch_table // output table for each patch
                   , const std::vector<float> &img, const unsigned width, const unsigned height, const unsigned chnls // input symetrized image
                   , const unsigned p, const unsigned q
                   , const unsigned K // Max number of other elements in output list for each input patch, the input patch is in the list
                   , const unsigned P, const unsigned Q // Size PxQ for the proximity of patches
                   , const float thresholdBM // Threshold value to keep or not a similar patch
                   , const std::vector<float> &sigma_table
                   , unsigned nb_threads
)
{
    //! PRECONDITION
    if (patch_table.size() != width * height)
        patch_table.resize(width * height);
#ifdef _OPENMP
    std::vector< omp_lock_t > lock_insert(patch_table.size());
    for (auto it = lock_insert.begin(); it != lock_insert.end() ; ++it)
        omp_init_lock(&(*it));
#endif
    //! Threshold
    float tauMatch = 0.0f;
    for (auto it = sigma_table.begin(); it != sigma_table.end(); ++it)
        tauMatch += (*it)*(*it);
    tauMatch *= 3.0f * thresholdBM;
    // Reserve to avoid reallocation in parallel writing
    for (auto t = patch_table.begin(); t != patch_table.end(); ++t)
    {
        t->reserve(K);
    }
    //! Hypothese
    //! The search start with the first patch in the central part of the input image
    int row = width + 2 * P;
    auto delta = Q * row + P;
    // 2] For all starting patches of all PxQ blocks  => OpenMP loop if enable
    unsigned int nPQ_blocks = width * height;
    float pf = static_cast<float>(p);
    float qf = static_cast<float>(q);
    float pqf = pf * qf;
    auto Pp = static_cast<int>(P-p);
    auto Qq = static_cast<int>(Q-q);
#ifdef _OPENMP
    omp_set_dynamic(0); omp_set_num_threads(nb_threads);
#endif
#pragma omp parallel for shared(patch_table) schedule(monotonic : static)
    for (unsigned int pIdx = 0u; pIdx < nPQ_blocks; pIdx++)
    {
        // iterator for the current patch
        unsigned int x = pIdx % width;
        unsigned int y = pIdx / width;
	// Prefering explicit casting even if useless
        int ix = static_cast<int>(x);
        int iy = static_cast<int>(y);
	int iw = static_cast<int>(width);
	int ih = static_cast<int>(height);
        unsigned int dec = delta + y * row + x;
        auto patch =  img.cbegin() + dec;
        //! Set of forward directions treated in sequential computation
        //! allowing dynamic values for p and q - not factorize before the loop !
        std::vector<int> dirs;
        int dir = - Pp - row * Qq;
        for (auto in_y = -Qq ; in_y <= Qq ; ++in_y, dir += row-(2*Pp+1))
        {
            for (auto in_x = -Pp ; in_x <= Pp ; ++in_x, dir++)
            {
//                dirs.push_back(dir);
		// Patch following dir is before the reference patch
		if ((in_y < 0) || ((in_y==0) && (in_x < 0))) {
			// Patch following dir is not an inner patch
			if (! (((0 <= ix+in_x) && (ix+in_x < iw)) \
			       && ((0 <= iy+in_y) && (iy+in_y < ih)) \
			      ) \
			   )
				dirs.push_back(dir);
		} else { // Patch following dir is after the reference patch
                        dirs.push_back(dir);
                } 
            }
        }
        // Only look for forward patches
        float dist = 0.0f;
        // For all directions
        for (auto d = dirs.begin(); d != dirs.end(); ++d)
        {
            // Compute similarity: for dynamic pq it is preferable to not incorporate pq in tauMatch formula
            dist = compute_dist(patch, patch + (*d), p, q, width, height, chnls, P, Q) / pqf;
            // Threshold to keep only valid patches
            if (dist < tauMatch)
            {
                unsigned absolute_center_dec = dec + (*d); // absolute displacement in image with boundary
                // Insert while preserving boundedness and heap property
                auto simPatch = std::make_pair(dist, absolute_center_dec);
#ifdef _OPENMP
                    omp_set_lock(&(lock_insert[pIdx]));
#endif
                insert_boundedSet(std::cref(simPatch), std::ref(patch_table[pIdx]), K);
#ifdef _OPENMP
                    omp_unset_lock(&(lock_insert[pIdx]));
#endif
                // Symmetrization of insert
                bool valid = true;
                auto sym_second_y = (absolute_center_dec / row);
                valid = valid && (sym_second_y >= Q);
                sym_second_y -= Q;
                valid = valid && (sym_second_y < height);
                auto sym_second_x = (absolute_center_dec % row);
                valid = valid && (sym_second_x >= P);
                sym_second_x -= P;
                valid = valid && (sym_second_x < width);
                // Test for a different inner patch avoiding putting twice the reference patch
                if (((*d) > 0) && (valid))
                { // the second patch is also in center image
                    auto pIdx_second = sym_second_y * width + sym_second_x;
                    auto neg_simPatch = std::make_pair(dist, dec);
                    // Possible data race: it can arrive with simultaneous insertion
                    // when only one slot is free so protect it.
#ifdef _OPENMP
                    omp_set_lock(&(lock_insert[pIdx_second]));
#endif
                    insert_boundedSet(std::cref(neg_simPatch), std::ref(patch_table[pIdx_second]), K);
#ifdef _OPENMP
                    omp_unset_lock(&(lock_insert[pIdx_second]));
#endif
                }
            }
        }
    }

#ifdef _OPENMP
    // Destroy lock when OpenMP is used
    for (auto it = lock_insert.begin(); it != lock_insert.end() ; ++it)
        omp_destroy_lock(&(*it));
#endif

}

