#include <cstdio>
#include <cstdlib>
#include <iostream>
#include <fstream>
#include <vector>
#include <cmath>
#include <cstring>

#include <Eigen/Dense>

#include "hosvd2.h"

/**
* @file   main.cpp
* @brief  Main executable file.
*
* @authors FABIEN FESCHET <fabien.feschet@dg-medical.fr> - 2018,
*
*/

/** ------------------ **/
/** - Check function - **/
/** ------------------ **/
/**
* @brief Check for the elements of the command-line
*
* @param argc and argv taken from the CLI of main
*
* @return EXIT_FAILURE if CLI is malformed, otherwise return EXIT_SUCCESS.
**/
int check(int argc, char**argv, Params &cli)
{
    bool malformed = false;
    for (int p = 1 ; p < argc ; ++p) {
            if (argv[p][0] == '-') {
                    if (argv[p][1] == 'i') {
                            cli.hasInput = true; cli.input = (char*)malloc(sizeof(char)*(1+strlen(argv[++p])));
                            strcpy(cli.input,argv[p]); continue;
                        }
                    if (argv[p][1] == 's') {
                            cli.hasSigma = true; cli.sigma = (float)atof(argv[++p]); continue;
                        }
                    if (argv[p][1] == 'S') {
                            cli.hasSigma = false; cli.sigma = (float)atof(argv[++p]); continue;
                        }
                    if (argv[p][1] == 'c') {
                            cli.hasColorspace = true; ++p;
                            cli.colorspace  = (strcmp(argv[p], "rgb"  ) == 0 ? RGB   :
                                              (strcmp(argv[p], "yuv"  ) == 0 ? YUV   :
                                              (strcmp(argv[p], "ycbcr") == 0 ? YCBCR :
                                              (strcmp(argv[p], "opp"  ) == 0 ? OPP   : NONE))));
                            continue;
                        }
                    // Test for output Parameters
                    if (argv[p][1] == 'o') {
                            if (argv[p][2] == 'n') {
                                    cli.out.hasOutputNoisy = true; cli.out.outputNoisy = (char*)malloc(sizeof(char)*(1+strlen(argv[++p])));
                                    strcpy(cli.out.outputNoisy,argv[p]); continue;
                                }
                            if (argv[p][2] == 'd' && argv[p][3] == '1') {
                                    cli.out.hasOutputDenoised1 = true; cli.out.outputDenoised1 = (char*)malloc(sizeof(char)*(1+strlen(argv[++p])));
                                    strcpy(cli.out.outputDenoised1,argv[p]); continue;
                                }
                            if (argv[p][2] == 'd' && argv[p][3] == '2') {
                                    cli.out.hasOutputDenoised2 = true; cli.out.outputDenoised2 = (char*)malloc(sizeof(char)*(1+strlen(argv[++p])));
                                    strcpy(cli.out.outputDenoised2,argv[p]); continue;
                                }
                            if (argv[p][2] == 'o') {
                                    cli.out.hasOutputOracle = true; cli.out.outputOracle = (char*)malloc(sizeof(char)*(1+strlen(argv[++p])));
                                    strcpy(cli.out.outputOracle,argv[p]); continue;
                                }
                        }
                    // Test for options Parameters
                    if (argv[p][1] == 'p') {
                            if (argv[p][2] == 'k') {
                                    cli.params.kHard = (unsigned)atoi(argv[++p]); continue;
                                }
                            if (argv[p][2] == 'K') {
                                    cli.params.K = (unsigned)atoi(argv[++p]); continue;
                                }
                            if (argv[p][2] == 'n') {
                                    cli.params.nHard = (unsigned)atoi(argv[++p]); continue;
                                }
                            if (argv[p][2] == 'B') {
                                    cli.params.thresholdBM = (float)atof(argv[++p]); continue;
                                }
                            if (argv[p][2] == 'H') {
                                    cli.params.thresholdHD = (float)atof(argv[++p]); continue;
                                }
                            if (argv[p][2] == 'p') {
                                    cli.params.pHard = (unsigned)atoi(argv[++p]); continue;
                                }
                            if (argv[p][2] == 't') {
                                    cli.params.nbThreads = (unsigned)atoi(argv[++p]); continue;
                                }
                        }
                } else {
                    std::cout << "...Malformed input detected at " << argv[p] << "..." << std::endl;
                    malformed = true;
                    break;
                }
        }
    // Check some variable to verify that the command-line is acceptable
    if (cli.hasInput == false || malformed) {
            std::cout << "Usage: HOSVDdenoising -i input [-s sigma || -S sigma] [-c colorspace]" << std::endl
                      << "                      [-on outputNoisy] [-od1 outputDenoised] [-od2 outputDenoisedWiener] [-oo outputOracle]" << std::endl
                      << "                      [-pk kHard] [-pK K] [-pn nHard] [-pB thresholdBlockMatching] [-pH thresholdHardThreshold] [-pt nb_threads]"
                      << std::endl;
            std::cout << "-i   input: input image. This parameter is mandatory." << std::endl
                      << "-s   sigma: add noise to the input image before denoising." << std::endl
                      << "-S   sigma: do not add noise to the input image but use sigma for denoising." << std::endl
                      << "-c   colorspace: choose between rgb, yuv, ycbcr, opp. Default is: rgb." << std::endl
                      << "-on  outputNoisy: to save the noisy image." << std::endl
                      << "-od1 outputDenoised: to save the denoised image without Wiener filtering." << std::endl
                      << "-od2 outputDenoisedWiener: to save the denoised image with Wiener filtering." << std::endl
                      << "-oo  outputOracle: to save the Oracle image." << std::endl
                      << "-pk  options: fix kHard." << std::endl
                      << "-pK  options: fix K." << std::endl
                      << "-pn  options: fix nHard." << std::endl
                      << "-pB  options: fix threshold factor for Block-Matching." << std::endl
                      << "-pH  options: fix threshold factor for Hard Thresholding." << std::endl
                      << "-pp  options: fix pHard the acceleration factor." << std::endl
                      << "-pt  options: fix number of OpenMP threads." << std::endl;
            return EXIT_FAILURE;
        }
    if (cli.hasColorspace && cli.colorspace == NONE) {
            std::cout << "warning: invalid colorspace detected. Options are: rgb, yuv, ycbcr or opp." << std::endl;
            std::cout << "........ Setting colorspace to default rgb." << std::endl;
            cli.colorspace = RGB;
        }

    return EXIT_SUCCESS;
}

/** ----------------- **/
/** - Main function - **/
/** ----------------- **/
/**
* @brief Main function
*
* @param argc, argv: command-line parameters
*
* @return EXIT_FAILURE if CLI is invalid otherwise return EXIT_SUCCESS.
**/
int main(int argc, char **argv)
{
    //! Check if there is the right call for the algorithm
    Params cli = {false,false,false,
                  NULL,RGB,0.0f,
                  {false,false,false,false,NULL,NULL,NULL,NULL},
                  {8,32,16,1.0f,1.0f,3,1}
                 };
    if (check(argc,argv,cli)==EXIT_FAILURE)
        return EXIT_FAILURE;

    //! Declarations
    std::vector<float> img, img_noisy, img_denoised, img_denoised_wiener, img_denoised_oracle;
    unsigned width, height, chnls;

    //! Load image
    if(load_image_IIO(cli.input, img, &width, &height, &chnls) != EXIT_SUCCESS)
        return EXIT_FAILURE;

    //! Add noise if necessary
    if (cli.hasSigma && cli.sigma > 0) {
            std::cout << std::endl << "Add noise [sigma = " << cli.sigma << "] ...";
            add_noise(img, img_noisy, cli.sigma);
            std::cout << "done." << std::endl;
        } else {
            std::cout << std::endl << "Not adding noise [but sigma = " << cli.sigma << "] ...";
            img_noisy = img;
        }

    //! Output noisy image in case of comparison with other algorithm
    if (cli.out.hasOutputNoisy) {
            std::cout << std::endl << "Save noisy image...";
            if (save_image(cli.out.outputNoisy, img_noisy, width, height, chnls) != EXIT_SUCCESS)
                return EXIT_FAILURE;
            std::cout << "done." << std::endl;
        }

    //! Denoising
    if (hosvd(cli.sigma, img, img_noisy, img_denoised, img_denoised_wiener, img_denoised_oracle,
                  width, height, chnls, cli.colorspace, cli.out, cli.params)!= EXIT_SUCCESS)
        return EXIT_FAILURE;

    //! Compute PSNR and RMSE
    float psnr, rmse;

    if (cli.out.hasOutputDenoised1) {
        if(compute_psnr(img, img_denoised, &psnr, &rmse) != EXIT_SUCCESS)
            return EXIT_FAILURE;

        std::cout << std::endl << "For 1st pass: ";
        std::cout << "PSNR: " << psnr << " / ";
        std::cout << "RMSE: " << rmse << std::endl;

        if (save_image(cli.out.outputDenoised1, img_denoised, width, height, chnls) != EXIT_SUCCESS)
            return EXIT_FAILURE;
    }

    if (cli.out.hasOutputDenoised2) {
            if(compute_psnr(img, img_denoised_wiener, &psnr, &rmse) != EXIT_SUCCESS)
                return EXIT_FAILURE;

            std::cout << std::endl << "For 2nd pass: ";
            std::cout << "PSNR: " << psnr << " / ";
            std::cout << "RMSE: " << rmse << std::endl;

            if (save_image_IIO(cli.out.outputDenoised2, img_denoised_wiener, width, height, chnls) != EXIT_SUCCESS)
                return EXIT_FAILURE;
        }
    if (cli.out.hasOutputOracle) {
            if(compute_psnr(img, img_denoised_oracle, &psnr, &rmse) != EXIT_SUCCESS)
                return EXIT_FAILURE;

            std::cout << std::endl << "For Oracle pass: ";
            std::cout << "PSNR: " << psnr << " / ";
            std::cout << "RMSE: " << rmse << std::endl;

            if (save_image(cli.out.outputOracle, img_denoised_oracle, width, height, chnls) != EXIT_SUCCESS)
                return EXIT_FAILURE;
        }

    //! End
    return EXIT_SUCCESS;
}
